unit ShapeObjDemomain;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, Menus, ComCtrls,
  ExtCtrls, ShapeObj, ShapeObjBGRA, WorldViewPortRoutines, WGS84, ClipperRoutines, StrUtils,
  ShapeTable, BGRAVirtualScreen, db, dbf, DBGrids, Buttons, ColorBox, StdCtrls,
  BGRABitmap, BCTypes, BGRAGraphicControl, BCPanel, uPoweredby, Math, Types;

type
  DBGridHack = class(TDBGrid);

  { TfrmShapeObjDemo }

  TfrmShapeObjDemo = class(TForm)
    BGRAGraphicControl1: TBGRAGraphicControl;
    btnDefault: TSpeedButton;
    btnPan: TSpeedButton;
    btnZoomIn: TSpeedButton;
    btnZoomOut: TSpeedButton;
    btnZoomToNormal: TSpeedButton;
    btnDrawShapeObj: TButton;
    cbxBrushColor: TColorBox;
    chkgrp1: TCheckGroup;
    cbxPenColor: TColorBox;
    Label1: TLabel;
    Label2: TLabel;
    lbledPenWidth: TLabeledEdit;
    lbledBufferWidth: TLabeledEdit;
    MainMenu1: TMainMenu;
    itemFile: TMenuItem;
    itemImportXYZTextData: TMenuItem;
    itemExport: TMenuItem;
    itemSimpleXYZFile: TMenuItem;
    itemEntityFormat: TMenuItem;
    itemShowShapeFiles: TMenuItem;
    itemAttributes: TMenuItem;
    itemExit: TMenuItem;
    itemTools: TMenuItem;
    itemDeleteShapeFiles: TMenuItem;
    itemCopyShapeFiles: TMenuItem;
    itemMergeShapeFiles: TMenuItem;
    itemDeleteShape: TMenuItem;
    itemCompressShapeFile: TMenuItem;
    OpenDialog1: TOpenDialog;
    Panel1: TPanel;
    Panel2: TPanel;
    Poweredby1: TPoweredby;
    ProgressBar1: TProgressBar;
    rgpShapeType: TRadioGroup;
    StatusBar1: TStatusBar;
    Timer1: TTimer;
    Timer2: TTimer;
    procedure BGRAGraphicControl1MouseDown(Sender: TObject;
      Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure BGRAGraphicControl1MouseMove(Sender: TObject; Shift: TShiftState;
      X, Y: Integer);
    procedure BGRAGraphicControl1MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure btnDefaultClick(Sender: TObject);
    procedure btnDrawShapeObjClick(Sender: TObject);
    procedure cbxPenColorChange(Sender: TObject);
    procedure chkgrp1ItemClick(Sender: TObject; Index: integer);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure itemAttributesClick(Sender: TObject);
    procedure itemCompressShapeFileClick(Sender: TObject);
    procedure itemCopyShapeFilesClick(Sender: TObject);
    procedure itemDeleteShapeClick(Sender: TObject);
    procedure itemDeleteShapeFilesClick(Sender: TObject);
    procedure itemEntityFormatClick(Sender: TObject);
    procedure itemExitClick(Sender: TObject);
    procedure itemImportXYZTextDataClick(Sender: TObject);
    procedure itemMergeShapeFilesClick(Sender: TObject);
    procedure itemShowShapeFilesClick(Sender: TObject);
    procedure itemSimpleXYZFileClick(Sender: TObject);
    procedure lbledBufferWidthChange(Sender: TObject);
    procedure lbledPenWidthChange(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure Timer2Timer(Sender: TObject);
  private
    FShapeObjBGRA : TShapeObjBGRA;
    FWVPortRtns : TWorldViewPortRoutines;
    FClipperRoutines : TClipperRoutines;
    FPenWidth, FBufferWidth : integer;
    FShapefileList, FSelectedShapesList : TStringList;
    Fxmin, Fymin, Fxmax, Fymax : double;
    FfrmShapeTable : TfrmShapeTable;
    FSelectedShapeNum : integer;
    FOpenMode : string;
    FMouseDown, FShapesLoaded : boolean;
    FZoomPanStatus :  integer;
    FWidth, FHeight : integer;
    FOrigin, FDestination : TPoint;
    FZoomRectangle : TRect;
    FAspectRatio : single;
    FPanMap : TBGRABitMap;

    procedure BufferShape(inWards: boolean);
    procedure DetermineMaxMinCoordinates;
    procedure DrawShapesShapeObj;
    procedure DrawZoomRectangle;
    procedure GetWorldCoordinates(x, y: integer; var xd, yd: double);
    procedure RestoreNormalZoom;
    procedure SetExtent;
    procedure SetPorts;
    procedure UpdateCoordinates;
  public
    procedure DrawShapes;
    procedure DrawSelectedShapes;
    procedure DrawShape(shapeNum: integer);
    procedure TwinkleShape(shapeNum: integer);

    property SelectedShapesList : TStringList read FSelectedShapesList
      write FSelectedShapesList;
    property ShapeObjBGRA : TShapeObjBGRA read FShapeObjBGRA
      write FShapeObjBGRA;
  end;

var
  frmShapeObjDemo: TfrmShapeObjDemo;

implementation

{$R *.lfm}

{ TfrmShapeObjDemo }

procedure TfrmShapeObjDemo.FormCreate(Sender: TObject);
begin
  Poweredby1.ShowPoweredByForm;

  FShapeObjBGRA := TShapeObjBGRA.Create(Self);
  FWVPortRtns := TWorldViewPortRoutines.Create(Self);
  FClipperRoutines := TClipperRoutines.Create(Self);
  FShapeObjBGRA.WorldViewPortRoutines := FWVPortRtns;
  FShapefileList := TStringList.Create;
  FSelectedShapesList := TStringList.Create;
  FSelectedShapesList.Duplicates := dupIgnore;
  FPanMap := TBGRABitmap.Create;

  FPenWidth := 1;
  FBufferWidth := 10;
  FOpenMode := 'rw';
  FMouseDown := false;
  FZoomPanStatus := 0;
  FShapesLoaded := false;
end;

procedure TfrmShapeObjDemo.FormDestroy(Sender: TObject);
begin
  FreeAndNil(FPanMap);
  FreeAndNil(FSelectedShapesList);
  FreeAndNil(FShapefileList);
end;

procedure TfrmShapeObjDemo.FormResize(Sender: TObject);
begin
  FWidth := BGRAGraphicControl1.Width;
  FHeight := BGRAGraphicControl1.Height;

  FPanMap.SetSize(FWidth, FHeight);

  FAspectRatio := Float(FHeight) / FWidth;

  Fxmin := FShapeObjBGRA.XMinTot - 50;
  Fymin := FShapeObjBGRA.YMinTot - 50;
  Fxmax := FShapeObjBGRA.XMaxTot + 50;
  Fymax := FShapeObjBGRA.YMaxTot + 50;

  SetPorts;

  Timer1.Enabled := true;
end;

procedure TfrmShapeObjDemo.Timer1Timer(Sender: TObject);
begin
  Timer1.Enabled := false;

  if FShapesLoaded then
    DrawShapes;
end;

procedure TfrmShapeObjDemo.Timer2Timer(Sender: TObject);
var
  n, shapeNum : integer;
begin
  Timer2.Enabled := false;

  n := 0;
  while n < FSelectedShapesList.Count do
  begin
    shapeNum := StrToInt(FSelectedShapesList.Strings[n]);
//    TwinkleShape(shapeNum);
    DrawShape(shapeNum);

    Inc(n);
  end;
end;

procedure TfrmShapeObjDemo.chkgrp1ItemClick(Sender: TObject; Index: integer);
begin
  if Index = 2 then
    FShapeObjBGRA.DrawNonDeletedShapesOnly := not chkgrp1.Checked[2]
  else if Index = 3 then
  begin
    chkgrp1.Checked[3] := false;
    FShapeObjBGRA.Compress;
  end
  else if Index = 4 then
  begin
    FZoomPanStatus := 0;
    BGRAGraphicControl1.Cursor := crDefault;
    Exit;
  end
  else if (Index = 5) or (Index = 6) then
  begin
    chkgrp1.Checked[Index] := false;
    MessageDlg('', 'Requires debugging', mtInformation, [mbOK], 0);
    Exit;
  end
  else if Index = 5 then
    BufferShape(false)
  else if Index = 6 then
    BufferShape(true);

  Timer1.Enabled := true;
end;

procedure TfrmShapeObjDemo.BGRAGraphicControl1MouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  n : integer;
  xd, yd : single;
begin
  FMouseDown := true;

  if FSelectedShapeNum = -1 then
  begin
    StatusBar1.SimpleText := '';
    StatusBar1.Font.Color := clBlack;
    xd := FWvPortRtns.wx(x);
    yd := FWvPortRtns.wy(y);
    FSelectedShapeNum := FShapeObjBGRA.HitTest(xd, yd);
    FShapeObjBGRA.DeleteShape(FSelectedShapeNum);
    FShapeObjBGRA.DrawNonDeletedShapesOnly := true;
    Timer1.Enabled := true;
    Exit;
  end;

  if chkgrp1.Checked[4] then
  begin
    xd := FWvPortRtns.wx(x);
    yd := FWvPortRtns.wy(y);
    n := FShapeObjBGRA.HitTest(xd, yd);
    FSelectedShapesList.Clear;
    FSelectedShapesList.Add(IntToStr(n));
    DrawSelectedShapes;
  end
  else
  begin
    if (FZoomPanStatus = 1) or (FZoomPanStatus = 2) then
    begin
      FOrigin.X := X; FOrigin.Y := Y;
      FDestination.X := X; FDestination.Y := Y;
      BGRAGraphicControl1.Bitmap.Canvas.Pen.Color := clBlack;
      BGRAGraphicControl1.Bitmap.Canvas.Pen.Mode := pmXOR;
      DrawZoomRectangle;
    end
    else if FZoomPanStatus = 3 then
    begin
      BGRAGraphicControl1.Bitmap.Draw(FPanMap.Canvas, 0, 0);
      FOrigin.X := X; FOrigin.Y := Y;
      StatusBar1.SimpleText :=
        'xd: ' +  IntToStr(Round(FWvPortRtns.wx(x))) + ' ' +
        'yd: ' +  IntToStr(Round(FWvPortRtns.wy(y))) + ' ' +
        'FOrigin.X: ' +  IntToStr(FOrigin.X) + ' ' +
        'FOrigin.Y: ' +  IntToStr(FOrigin.Y);
    end;
  end;
end;

procedure TfrmShapeObjDemo.BGRAGraphicControl1MouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var
  xdiff, ydiff : integer;
begin
  if (FMouseDown) and ((FZoomPanStatus = 1) or (FZoomPanStatus = 2)) then
  begin
    DrawZoomRectangle;
    FDestination.X := X; FDestination.Y := Y;
    DrawZoomRectangle;
  end
  else if (FMouseDown) and (FZoomPanStatus = 3) then
  begin
    BGRAGraphicControl1.Bitmap.CanvasBGRA.Brush.Color := clWhite;
    BGRAGraphicControl1.Bitmap.CanvasBGRA.FillRect(0, 0, FWidth, FHeight);
    FDestination.X := X; FDestination.Y := Y;
    xdiff := FDestination.X - FOrigin.X;
    ydiff := FDestination.Y - FOrigin.Y;
{    StatusBar1.SimpleText :=
      'xd: ' +  IntToStr(Round(FWvPortRtns.wx(x))) + ' ' +
      'yd: ' +  IntToStr(Round(FWvPortRtns.wy(y))) + ' ' +
      'FOrigin.X: ' +  IntToStr(FOrigin.X) + ' ' +
      'FOrigin.Y: ' +  IntToStr(FOrigin.Y) + ' ' +
      'FDestination.X: ' +  IntToStr(FDestination.X) + ' ' +
      'FDestination.Y: ' +  IntToStr(FDestination.Y) + ' ' +
      'xdiff: ' +  IntToStr(Round(xdiff)) + ' ' +
      'ydiff: ' +  IntToStr(Round(ydiff));}
    FPanMap.Draw(BGRAGraphicControl1.Bitmap.Canvas, xdiff, ydiff);
    BGRAGraphicControl1.Refresh;
    Exit;
  end;

{  StatusBar1.SimpleText :=
    'xd: ' +  IntToStr(Round(FWvPortRtns.wx(x))) + ' ' +
    'yd: ' +  IntToStr(Round(FWvPortRtns.wy(y))) + ' ' +
    'FOrigin.X: ' +  IntToStr(FOrigin.X) + ' ' +
    'FOrigin.Y: ' +  IntToStr(FOrigin.Y);}
end;

procedure TfrmShapeObjDemo.BGRAGraphicControl1MouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  FMouseDown := false;

  if (FZoomPanStatus = 1) or (FZoomPanStatus = 2) then
  begin
    BGRAGraphicControl1.Cursor := crDefault;
    FDestination.X := X; FDestination.Y := Y;
    DrawZoomRectangle;
    BGRAGraphicControl1.Bitmap.Canvas.Pen.Mode := pmCopy;

    SetExtent;
  end
  else if FZoomPanStatus = 3 then
  begin
    FDestination.X := X; FDestination.Y := Y;
    UpdateCoordinates;
  end;
end;

procedure TfrmShapeObjDemo.DrawZoomRectangle;
var
  res : string;
begin
  FZoomRectangle.Left := Min(FOrigin.x, FDestination.x);
  FZoomRectangle.Top := Min(FOrigin.y, FDestination.y);
  FZoomRectangle.Right := Max(FOrigin.x, FDestination.x);
  FZoomRectangle.Bottom := Max(FOrigin.y, FDestination.y);

  BGRAGraphicControl1.Bitmap.Canvas.DrawFocusRect(FZoomRectangle);
  BGRAGraphicControl1.Refresh;
end;

procedure TfrmShapeObjDemo.UpdateCoordinates;
var
  x1, y1, x2, y2 : integer;
  xd1, yd1, xd2, yd2, xwdiff, ywdiff : single;
  xmin, ymin, xmax, ymax : single;
begin
  xmin := FWvPortRtns.xmin;
  ymin := FWvPortRtns.ymin;
  xmax := FWvPortRtns.xmax;
  ymax := FWvPortRtns.ymax;

  xd1 := FWvPortRtns.wx(FOrigin.X);
  yd1 := FWvPortRtns.wy(FOrigin.Y);
  xd2 := FWvPortRtns.wx(FDestination.X);
  yd2 := FWvPortRtns.wy(FDestination.Y);

  xwdiff := xd2 - xd1;
  ywdiff := yd2 - yd1;

  xd1 := xmin - xwdiff;
  yd1 := ymin - ywdiff;
  xd2 := xmax - xwdiff;
  yd2 := ymax - ywdiff;

  if xd1 = xd2 then
  begin
    xd1 := FWvPortRtns.xmin;
    xd2 := FWvPortRtns.xmax;
  end;

  if yd1 = yd2 then
  begin
    yd1 := FWvPortRtns.ymin;
    yd2 := FWvPortRtns.ymax;
  end;

  Fxmin := xd1; Fymin := yd1;
  Fxmax := xd2; Fymax := yd2;

  SetPorts;

  Timer1.Enabled := true;
end;

procedure TfrmShapeObjDemo.SetExtent;
var
  xd1, yd1, xd2, yd2, xc, yc, xwdiff, xwdiffHalf, xdiff : single;
  wxmin, wxmax, dFactor : single;
begin
  wxmin := FWvPortRtns.xmin;
  wxmax := FWvPortRtns.xmax;

  xd1 := FWvPortRtns.wx(FZoomRectangle.Left);
  yd1 := FWvPortRtns.wy(FZoomRectangle.Bottom);
  xd2 := FWvPortRtns.wx(FZoomRectangle.Right);
  yd2 := FWvPortRtns.wy(FZoomRectangle.Top);

  xc := (xd2 + xd1) / 2;
  yc := (yd2 + yd1) / 2;

  xdiff := FZoomRectangle.Right - FZoomRectangle.Left;
  if xdiff = 0 then Exit;

  if FZoomPanStatus = 1 then
    xwdiff := xd2 - xd1
  else
  begin
    dFactor := Float(FWidth) / xdiff;
    xwdiff := (wxmax - wxmin) * dFactor;
  end;

  xwdiffHalf := xwdiff / 2;

  xd1 := xc - xwdiffHalf;
  yd1 := yc - xwdiffHalf * FAspectRatio;
  xd2 := xc + xwdiffHalf;
  yd2 := yc + xwdiffHalf * FAspectRatio;

  Fxmin := xd1; Fymin := yd1;
  Fxmax := xd2; Fymax := yd2;

  SetPorts;

  Timer1.Enabled := true;
end;

procedure TfrmShapeObjDemo.btnDefaultClick(Sender: TObject);
begin
  chkgrp1.Checked[4] := false;

  FZoomPanStatus := TSpeedButton(Sender).Tag;
  case FZoomPanStatus of
    0:
    begin
      BGRAGraphicControl1.Cursor := crDefault;
    end;
    1, 2: BGRAGraphicControl1.Cursor := crCross;
    3:
    begin
      BGRAGraphicControl1.Cursor := crHandPoint;
    end;
    4:
    begin
      BGRAGraphicControl1.Cursor := crDefault;
      RestoreNormalZoom;
    end;
  end;
end;

procedure TfrmShapeObjDemo.cbxPenColorChange(Sender: TObject);
begin
  Timer1.Enabled := true;
end;

procedure TfrmShapeObjDemo.RestoreNormalZoom;
begin
  Fxmin := FShapeObjBGRA.XMinTot - 50;
  Fymin := FShapeObjBGRA.YMinTot - 50;
  Fxmax := FShapeObjBGRA.XMaxTot + 50;
  Fymax := FShapeObjBGRA.YMaxTot + 50;

  SetPorts;

  Timer1.Enabled := true;
end;

procedure TfrmShapeObjDemo.SetPorts;
begin
  FWvPortRtns.SetViewPort(0, 0, FWidth, FHeight);
  FWvPortRtns.SetWorldPort(
    Fxmin, Fymin, Fxmax, Fymax);
end;

procedure TfrmShapeObjDemo.GetWorldCoordinates(x, y : integer;
  var xd, yd : double);
begin
  xd := FWVPortRtns.wx(x);
  yd := FWVPortRtns.wy(y);
end;

procedure TfrmShapeObjDemo.itemExitClick(Sender: TObject);
begin
  Close;
end;

procedure TfrmShapeObjDemo.itemImportXYZTextDataClick(Sender: TObject);
var
  shapeFileName, line, res : string;
  tempList, tempList2 : TStringList;
  n, numValues, numPoints : integer;
  pointDataType : boolean;
  x, y, z : single;
  shapeType : integer;

  function ExistsForPossibleOldShapeFileWithSameName : boolean;
  begin
    Result := true;
    // check for possible already existing shapefile with current name
    if FShapeObjBGRA.Exists(shapeFileName) then
    begin
      if MessageDlg('', ExtractFilename(shapeFileName) +
        ' already exist' + lineEnding + 'OK to overwrite', mtConfirmation,
        [mbOK, mbNo], 0, mbNo) = mrNo then Exit;

      res := FShapeObjBGRA.DeleteShapeFile(shapeFileName);
      if res <> 'Done' then
      begin
        ShowMessage('Did not succeed to delete old shapefile:' + lineEnding +
          shapeFileName + lineEnding +
          'maybe open in some other program');
        Exit;
      end;
    end;

    Result := false;
  end;

  procedure RemovePossibleTabsCommasAndDoubleSpaces;
  begin
    repeat
      line := AnsiReplaceText(line, ',', ' ');
    until PosEx(chr(9) + chr(9), line) <= 0;
    repeat
      line := AnsiReplaceText(line, chr(9) + chr(9), ' ');
    until PosEx(chr(9) + chr(9), line) <= 0;
    repeat
      line := AnsiReplaceText(line, chr(9), ' ');
    until PosEx(chr(9), line) <= 0;
    repeat
      line := AnsiReplaceText(line, '  ', ' ');
    until PosEx('  ', line) <= 0;
    line := TrimRight(TrimLeft(line));
  end;

  procedure CreateShapeFile;
  var
    n : integer;

    function GetNumberOfValues : integer;
    begin
      line := tempList.Strings[0];
      RemovePossibleTabsCommasAndDoubleSpaces;
      tempList2.Text := AnsiReplaceText(line, ' ', lineEnding);
      Result := tempList2.Count;
    end;

  begin
    numValues := GetNumberOfValues;

    case rgpShapeType.ItemIndex of
      0:
      begin
        if numValues = 2 then
          shapeType := 1
        else
          shapeType := 11;
      end;
      1:
      begin
        if numValues = 2 then
          shapeType := 3
        else
          shapeType := 13;
      end;
      2:
      begin
        if numValues = 2 then
          shapeType := 5
        else
          shapeType := 15;
      end;
    end;

    FShapeObjBGRA.CreateShapeFile(shapeFileName, shapeType)
  end;

  function ExtractXYZ : boolean;
  begin
    RemovePossibleTabsCommasAndDoubleSpaces;
    if PosEx('X', UpperCase(line)) <= 0 then
    begin
      tempList2.Text := AnsiReplaceText(line, ' ', lineEnding);
      x := StrToFloat(TrimRight(TrimLeft(tempList2.Strings[0])));
      y := StrToFloat(TrimRight(TrimLeft(tempList2.Strings[1])));
      if numValues = 3 then
        z := StrToFloat(TrimRight(TrimLeft(tempList2.Strings[2])))
      else
        z := 0;
      Result := true;
    end
    else
      Result := false;
  end;

begin
  OpenDialog1.FileName := '*.txt';
  if not OpenDialog1.Execute then Exit;

  shapeFileName := ChangeFileExt(OpenDialog1.FileName, '');
  if not InputQuery('', 'Enter shapefilename', shapeFileName) then Exit;

  if ExistsForPossibleOldShapeFileWithSameName then Exit;

  pointDataType := rgpShapeType.ItemIndex = 0;

  tempList := TStringList.Create;
  tempList2 := TStringList.Create;
  try
    tempList.LoadFromFile(OpenDialog1.FileName);

    // create shapefile
    CreateShapeFile;

    // open shapefile for readwrite
    FShapeObjBGRA.Open(shapeFileName, 'rw');

    n := 0;
    while n < tempList.Count do
    begin
      line := TrimRight(TrimLeft(tempList.Strings[n]));
      if PosEx('X', UpperCase(line)) <= 0 then
      begin
        if (not pointDataType) and ((n = 0) or (not ExtractXYZ)) then
        begin
          if FShapeObjBGRA.NumPoints > 0 then
            FShapeObjBGRA.ClosePolygonPolylineShape; // close current shape
          FShapeObjBGRA.AddPolygonPolylineShape;
        end
        else
          FShapeObjBGRA.Add(x, y, z, 0);
      end
      else
      begin
        if FShapeObjBGRA.NumPoints > 0 then
          FShapeObjBGRA.ClosePolygonPolylineShape; // close current shape
        FShapeObjBGRA.AddPolygonPolylineShape;
      end;

      Inc(n);
    end;

    if not pointDataType then
      FShapeObjBGRA.ClosePolygonPolylineShape; // close last shape
    FShapeObjBGRA.Close;
  finally
    FreeAndNil(tempList2);
    FreeAndNil(tempList);
  end;
end;

procedure TfrmShapeObjDemo.itemEntityFormatClick(Sender: TObject);
var
  dir, namn : string;
begin
  if not FShapeObjBGRA.Select(FOpenMode, 'Select shapefile') then Exit;

  dir := FShapeObjBGRA.ShapeFileDir;
  namn := ExtractFileName(FShapeObjBGRA.ShapeFileName);
  FShapeObjBGRA.ExportAsTextFile(dir + namn, dir + namn, -1,
    ProgressBar1, StatusBar1);
end;

procedure TfrmShapeObjDemo.itemSimpleXYZFileClick(Sender: TObject);
var
  dir, namn : string;
begin
  if not FShapeObjBGRA.Select(FOpenMode, 'Select shapefile') then Exit;

  dir := FShapeObjBGRA.ShapeFileDir;
  namn := ExtractFileName(FShapeObjBGRA.ShapeFileName);
  FShapeObjBGRA.ExportAsXYZMTextFile(dir + namn, dir + namn, -1, true,
    FShapeObjBGRA.xmin, FShapeObjBGRA.ymin,
    FShapeObjBGRA.xmax, FShapeObjBGRA.ymax,
    ProgressBar1, StatusBar1);
end;

procedure TfrmShapeObjDemo.lbledBufferWidthChange(Sender: TObject);
var
  w, i : integer;
begin
  Val(lbledBufferWidth.Text, w, i);
  if i = 0 then
    FBufferWidth := w
  else
    lbledBufferWidth.Text := IntToStr(FBufferWidth);

  Timer1.Enabled := true;
end;

procedure TfrmShapeObjDemo.lbledPenWidthChange(Sender: TObject);
var
  w, i : integer;
begin
  Val(lbledPenWidth.Text, w, i);
  if i = 0 then
    FPenWidth := w
  else
    lbledPenWidth.Text := IntToStr(FPenWidth);

  Timer1.Enabled := true;
end;

procedure TfrmShapeObjDemo.itemShowShapeFilesClick(Sender: TObject);
var
  n : integer;
begin
  FShapefileList.Clear;
  if not FShapeObjBGRA.MultiSelect('Select shapefile(s)', FShapefileList) then Exit;

  DetermineMaxMinCoordinates;
  Caption := 'ShapeObjDemo';
  if FShapefileList.Count = 1 then
    Caption := Caption + ' - ' + FShapeObjBGRA.ShapeFileName;

  // Set viewport to be able to show world-coordinates
  FWVPortRtns.SetViewPort(0, 0,
    BGRAGraphicControl1.Width, BGRAGraphicControl1.Height);
  FWVPortRtns.SetWorldPort(Fxmin, Fymin, Fxmax, Fymax);

  FShapesLoaded := true;

  Timer1.Enabled := true;
end;

procedure TfrmShapeObjDemo.DrawShapes;
var
  n, i, shapeNum : integer;
  tempList : TStringList;
begin
  BGRAGraphicControl1.Bitmap.CanvasBGRA.Brush.Color := clWhite;
  BGRAGraphicControl1.Bitmap.CanvasBGRA.FillRect(0, 0, FWidth, FHeight);

  tempList := TStringList.Create;
  try
    n := 0;
    while n < FShapefileList.Count do
    begin
      FShapeObjBGRA.Open(FShapefileList.Strings[n], FOpenMode);

      if FShapeObjBGRA.SelectShapesWithinRectangel(
        Fxmin, Fymin, Fxmax, Fymax, tempList) then
      begin
        i := 0;
        while i < tempList.Count do
        begin
          shapeNum := StrToInt(tempList.Strings[i]);
          DrawShape(shapeNum);

          Inc(i);

          if i mod 1000 = 0 then
            BGRAGraphicControl1.Refresh;
        end;
      end;
      BGRAGraphicControl1.Refresh;

      Inc(n);
    end;
  finally
    FreeAndNil(tempList);
  end;
end;

procedure TfrmShapeObjDemo.DrawSelectedShapes;
var
  n, shapeNum : integer;
  xd, yd : single;
begin
  shapeNum := StrToInt(FSelectedShapesList.Strings[0]);

  if FfrmShapeTable = nil then
    FfrmShapeTable := TfrmShapeTable.Create(Self);

  FfrmShapeTable.DisplayTable(FShapeObjBGRA.ShapeFileDir +
    ChangeFileExt(FShapeObjBGRA.ShapeFileName, '.dbf'));
  FfrmShapeTable.Show;
  FfrmShapeTable.Dbf1.First;
  FfrmShapeTable.Dbf1.MoveBy(shapeNum);

  FShapeObjBGRA.ReadShape(shapeNum);

  xd := FShapeObjBGRA.XMax - FShapeObjBGRA.XMin;
  yd := FShapeObjBGRA.YMax - FShapeObjBGRA.YMin;

  Fxmin := Max(FShapeObjBGRA.XMinTot - 50, FShapeObjBGRA.XCenter - 10 * xd);
  Fymin := Max(FShapeObjBGRA.YMinTot - 50, FShapeObjBGRA.YCenter - 10 * yd);
  Fxmax := Min(FShapeObjBGRA.XMaxTot + 50, FShapeObjBGRA.XCenter + 10 * xd);
  Fymax := Min(FShapeObjBGRA.YMaxTot + 50, FShapeObjBGRA.YCenter + 10 * yd);

  SetPorts;

  DrawShapes;
  TwinkleShape(shapeNum);
  DrawShape(shapeNum);
end;

procedure TfrmShapeObjDemo.btnDrawShapeObjClick(Sender: TObject);
begin
  FShapeObjBGRA.Close;
  FShapefileList.Clear;
  if not FShapeObjBGRA.MultiSelect('Select shapefile(s)', FShapefileList) then Exit;
  DrawShapesShapeObj;
end;

procedure TfrmShapeObjDemo.DrawShapesShapeObj;
var
  n, i, shapeNum : integer;
  tempList : TStringList;
  Ellapsed1, Ellapsed2, Ellapsed3 : LongWord;
  ShapeObj : TShapeObj;
begin
{  FShapeObjBGRA.Close;

  BGRAGraphicControl1.Bitmap.CanvasBGRA.Brush.Color := clWhite;
  BGRAGraphicControl1.Bitmap.CanvasBGRA.FillRect(0, 0, FWidth, FHeight);

  ShapeObj := TShapeObj.Create(nil);
  tempList := TStringList.Create;
  try
    ShapeObj.Canvas := PaintBox1.Canvas;
    n := 0;
    while n < FShapefileList.Count do
    begin
      ShapeObj.Open(FShapefileList.Strings[n], FOpenMode);
      Ellapsed1 := GetTickCount;
      ShapeObj.DrawShapes;
      Ellapsed1 := GetTickCount - Ellapsed1;

      Ellapsed2 := GetTickCount;
      shapeNum := 0;
      while shapeNum < ShapeObj.NumEntities do
      begin
        ShapeObj.DrawShape(PaintBox1.Canvas, shapeNum);

        Inc(shapeNum);
      end;
      Ellapsed2 := GetTickCount - Ellapsed2;

      Inc(n);
    end;
    StatusBar1.SimpleText :=
    'Ellapsed1: ' + IntToStr(Ellapsed1) + ' ' +
    'Ellapsed2: ' + IntToStr(Ellapsed2);
  finally
    FreeAndNil(tempList);
    FreeAndNil(ShapeObj);
  end;}
end;

procedure TfrmShapeObjDemo.TwinkleShape(shapeNum : integer);
begin
  FShapeObjBGRA.TwinkleShape(cbxBrushColor.Selected, shapeNum, 2);
end;

procedure TfrmShapeObjDemo.DrawShape(shapeNum : integer);
begin
  FShapeObjBGRA.DrawShape(BGRAGraphicControl1, shapeNum, FPenWidth,
    cbxPenColor.Selected, cbxBrushColor.Selected,
    chkgrp1.Checked[0], chkgrp1.Checked[1]);
end;

procedure TfrmShapeObjDemo.DetermineMaxMinCoordinates;
var
  n : integer;
begin
  // determine total extent
  Fxmin := 100000000; Fymin := 100000000;
  Fxmax := -100000000; Fymax := -100000000;
  n := 0;
  while n < FShapefileList.Count do
  begin
    FShapeObjBGRA.Open(FShapefileList.Strings[n], FOpenMode);
    if Fxmin > FShapeObjBGRA.XMinTot then
      Fxmin := FShapeObjBGRA.XMinTot;
    if Fymin > FShapeObjBGRA.YMinTot then
      Fymin := FShapeObjBGRA.YMinTot;
    if Fxmax < FShapeObjBGRA.XMaxTot then
      Fxmax := FShapeObjBGRA.XMaxTot;
    if Fymax < FShapeObjBGRA.YMaxTot then
      Fymax := FShapeObjBGRA.YMaxTot;

    Inc(n);
  end;

  Fxmin := Fxmin - 50;
  Fymin := Fymin - 50;
  Fxmax := Fxmax + 50;
  Fymax := Fymax + 50;
end;

procedure TfrmShapeObjDemo.itemAttributesClick(Sender: TObject);
begin
  if not FShapesLoaded then
  begin
    MessageDlg('First open shapefile', mtInformation, [mbOK], 0);
    Exit;
  end;

  if FfrmShapeTable = nil then
    FfrmShapeTable := TfrmShapeTable.Create(Self);

  FfrmShapeTable.DisplayTable(FShapeObjBGRA.ShapeFileDir +
    ChangeFileExt(FShapeObjBGRA.ShapeFileName, '.dbf'));
end;

procedure TfrmShapeObjDemo.itemCopyShapeFilesClick(Sender: TObject);
var
  ShapeFileName, ShapeFileCopyName, dir : string;
begin
  if not FShapeObjBGRA.Select(FOpenMode, 'Select shapefile') then Exit;

  dir := FShapeObjBGRA.ShapeFileDir;
  ShapeFileName := FShapeObjBGRA.ShapeFileName;
  ShapeFileCopyName := ChangeFileExt(ShapeFileName, '_Copy');
  if not InputQuery('', 'Enter shapefile copyname', ShapeFileCopyName) then Exit;

  FShapeObjBGRA.Close;
  FShapeObjBGRA.CopyShapeFile(dir + ShapeFileName, dir + ShapeFileCopyName);
end;

procedure TfrmShapeObjDemo.itemDeleteShapeClick(Sender: TObject);
begin
  if not FShapeObjBGRA.Active then Exit;

  FSelectedShapeNum := -1;
  StatusBar1.SimpleText := 'SELECT SHAPE TO DELETE';
end;

procedure TfrmShapeObjDemo.itemDeleteShapeFilesClick(Sender: TObject);
var
  n : integer;
begin
  FShapefileList.Clear;
  if not FShapeObjBGRA.MultiSelect('Select shapefile(s)', FShapefileList) then Exit;

  if FShapeObjBGRA.Active then
    FShapeObjBGRA.Close;

  n := 0;
  while n < FShapefileList.Count do
  begin
    FShapeObjBGRA.DeleteShapeFile(FShapefileList.Strings[n]);

    Inc(n);
  end;
end;

procedure TfrmShapeObjDemo.itemMergeShapeFilesClick(Sender: TObject);
begin
  FShapeObjBGRA.MergeShapeFiles;
end;

procedure TfrmShapeObjDemo.itemCompressShapeFileClick(Sender: TObject);
begin
  if not FShapeObjBGRA.Select(FOpenMode, 'Select shapefile to compress') then Exit;

  FfrmShapeTable.Dbf1.Active := false;

  FShapeObjBGRA.Compress;

  DrawShapes;
end;

procedure TfrmShapeObjDemo.BufferShape(inWards: boolean);
var
  shapeObj, bufferedShape : TShapeObj;
  bufferShapefileName, dir, res : string;
begin
  shapeObj := TShapeObj.Create(nil);
  bufferedShape := TShapeObj.Create(nil);
  try
    if FShapeObjBGRA.Active then
      FShapeObjBGRA.Close;

    if not shapeObj.Select(FOpenMode, 'Select shapefile to buffer') then Exit;

    dir := shapeObj.ShapeFileDir;
    bufferShapefileName := dir +
      ChangeFileExt(shapeObj.ShapeFileName, '_buffered.shp');
    if shapeObj.Exists(bufferShapefileName) then
      shapeObj.DeleteShapeFile(bufferShapefileName);
    bufferedShape.CreateShapeFile(bufferShapefileName, shapeObj.ShapeType);
    bufferedShape.Open(bufferShapefileName, 'rw');
    res := FClipperRoutines.BufferShape(shapeObj, inWards,
             FBufferWidth, bufferedShape);
    bufferedShape.Close;
    shapeObj.Close;
  finally
    FreeAndNil(bufferedShape);
    FreeAndNil(shapeObj);
  end;
end;

end.

